// This file is part of Lanradio.
//
// Lanradio is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Lanradio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Lanradio.  If not, see <http://www.gnu.org/licenses/>.
//
// Authors:
// Florian Klingler <klingler@ccs-labs.org>

#include <stdio.h>
#include <stdlib.h>
#include "zmq/zmq.hpp"
#include <string>
#include <iostream>
#include <thread>
#include "pstreams/pstream.h"
#include "wlan.pb.h"

#include "SendToHardwareViaMAC.cc"
#include "ReceiveFromHardwareViaMAC.cc"
#include <boost/program_options.hpp>


#define DEBUG 0


namespace po = boost::program_options;

//  recv zmq string
static std::string myzmq_recv(zmq::socket_t &sock) {

    zmq::message_t msg;
    sock.recv(&msg);

    return std::string(static_cast<char *>(msg.data()), msg.size());
}

//  send zmq string
static bool myzmq_send(zmq::socket_t &sock, const std::string &str) {

    zmq::message_t msg(str.size());
    memcpy(msg.data(), str.data(), str.size());

    bool ret = sock.send(msg);
    return ret;
}

//  send zmq string, sendmore
static bool myzmq_sndmore(zmq::socket_t &sock, const std::string &str) {

    zmq::message_t msg(str.size());
    memcpy(msg.data(), str.data(), str.size());

    bool ret = sock.send(msg, ZMQ_SNDMORE);
    return ret;
}

void configure_radio_thread(int radioconfig_zmq_port) {

    std::string radioconfig_zmq_hostname = "*";
    zmq::context_t* radioconfig_zmq_context = new zmq::context_t(1);
    zmq::socket_t* radioconfig_zmq_socket= new zmq::socket_t(*radioconfig_zmq_context, ZMQ_REP);

    radioconfig_zmq_socket->bind(("tcp://" + radioconfig_zmq_hostname + ":" + std::to_string(radioconfig_zmq_port)).c_str());

    while (true) {
        std::string req = myzmq_recv(*radioconfig_zmq_socket);

        //        if(
        //                req.find("ifconfig ") == 0 ||
        //                req.find("iw ") == 0) {
        //        } else {
        //            std::cerr << "invalid cmd: " << req << std::endl;
        //            myzmq_send(*radioconfig_zmq_socket, "openc2x: invalid command");
        //            continue;
        //        }

        if(DEBUG) std::cout << "radioconfig\n\trcvd: " << req << std::endl;

        // run a process and create a streambuf that reads its stdout and stderr
        redi::ipstream proc(req, redi::pstreams::pstdout | redi::pstreams::pstderr);
        proc.close();

        int status = proc.rdbuf()->status();
        int exit_status = WEXITSTATUS(status);
        //std::cerr << "exit status: " << std::to_string(WEXITSTATUS(exit_status)) << std::endl;

        std::stringstream resp_stdout;
        std::stringstream resp_stderr;

        resp_stdout << proc.out().rdbuf();
        resp_stderr << proc.err().rdbuf();

        myzmq_send(*radioconfig_zmq_socket, resp_stdout.str());
        //myzmq_send(*radioconfig_zmq_socket, std::to_string(exit_status));
    }
}

void tx_wlan_thread(SendToHardwareViaMAC* mac, int messages_zmq_port) {

    std::string messages_envelope = "";

    std::string messages_zmq_hostname = "localhost";
    zmq::context_t* messages_zmq_context = new zmq::context_t(1);
    zmq::socket_t* messages_zmq_socket= new zmq::socket_t(*messages_zmq_context, ZMQ_SUB);

    messages_zmq_socket->connect(("tcp://" + messages_zmq_hostname + ":" + std::to_string(messages_zmq_port)).c_str());
    if (messages_envelope == "") {
        messages_zmq_socket->setsockopt(ZMQ_SUBSCRIBE, "", 0);//subscribe to all messages
    } else {
        messages_zmq_socket->setsockopt(ZMQ_SUBSCRIBE, messages_envelope.c_str(), 1);
    }

    while(true) {

        std::string envelope = myzmq_recv(*messages_zmq_socket);
        std::string message = myzmq_recv(*messages_zmq_socket);

        //std::cerr << "received from veins: length=" << message.length() << "\n\t" << "envelope: " << envelope << "\n\t" << "msg: " << message << std::endl;
        if (DEBUG) std::cerr << "received from veins: " << "\n\t" << "envelope: " << envelope << "\n\t" << "msg-length: " << message.length() << std::endl << std::endl;;

        if(envelope=="WLAN") {
            wlanPackage::WLAN* w = new wlanPackage::WLAN();
            bool succesful = w->ParseFromString(message);

            if(succesful) {
                string payload = w->payload();
                string srcmac = w->srcmac();
                string dstmac = w->dstmac();
                uint16_t ether_type = (uint16_t) w->ethertype();

                if (DEBUG) std::cerr << "payload-length = " << payload.size() << std::endl;

                mac->sendWLAN(&payload, PRIORITY_BE, &srcmac, &dstmac, &ether_type);
            }
        }
    }
}



void rx_wlan_thread(ReceiveFromHardwareViaMAC* mac_rx, SendToHardwareViaMAC* mac_tx, int portOut, bool receive_own_msgs, bool receive_every_ether_type) {

    zmq::context_t* mContext = new zmq::context_t(1);
    zmq::socket_t* mPublisher = new zmq::socket_t(*mContext, ZMQ_PUB);
    mPublisher->bind(("tcp://*:" + std::to_string(portOut)).c_str());

    pair<ReceivedPacketInfo, string> receivedData;      //MAC Sender, serialized DATA
    ReceivedPacketInfo* pktInfo = &receivedData.first;
    string* serializedData = &receivedData.second;

    while (true) {
        receivedData = mac_rx->receiveWLAN(receive_own_msgs, receive_every_ether_type);  //receive serialized DATA


        //check whether the mac of the sender and our own mac are the same and discard the package if we want to ignore those packages
        //if(!should_receive_own_msgs && /*(pktInfo->mSenderMac.compare(mac_tx->mOwnMac) == 0)*/ pktInfo->was_outgoing_packet) {
        //
        //    if(DEBUG) std::cerr << "received own Message, discarding" << std::endl;
        //    continue;
        //}

        if(DEBUG) std::cerr << "received from HW: " << "\n\t" << "msg-length: " << serializedData->length() << std::endl << std::endl;


        wlanPackage::WLAN wlan;
        wlan.set_payload(*serializedData);
        wlan.set_srcmac(pktInfo->mSenderMac);
        wlan.set_dstmac(pktInfo->mReceiverMac);
        wlan.set_priority(wlanPackage::WLAN_Priority_NA);
        wlan.set_ethertype((uint32_t) pktInfo->ether_type);
        wlan.set_isloopback(pktInfo->was_outgoing_packet);

        if(DEBUG) std::cerr << "payload-length (rx) = " << wlan.payload().size() << std::endl;

        std::string ser;
        wlan.SerializeToString(&ser);

        myzmq_sndmore(*mPublisher, "WLAN");
        myzmq_send(*mPublisher, ser);
    }
}




int main(int ac, char* av[]) {

    std::string interface;
    bool echo;
    int baseport;

    try {

        po::options_description desc("Allowed options");
        desc.add_options()

                                ("help, h", "show options")
                                ("interface, i", po::value<std::string>(&interface)->default_value("wlan0"), "wlan interface")
                                ("echo, e", po::value<bool>(&echo)->default_value(false), "send copy of all transmitted packets back to sender, similar to monitor mode (without radiotap information)")
                                ("baseport, p", po::value<int>(&baseport)->default_value(19191), "radioconfig = baseport; tx_wlan = baseport+1, rx_wlan = baseport+2")
                        ;

        po::variables_map vm;
        po::store(po::parse_command_line(ac, av, desc), vm);
        po::notify(vm);

        if (vm.count("help")) {
            cout << desc << "usage: --interface wlan0 --echo 1 --baseport 19191\n";
            return 0;
        }

    }
    catch(exception& e) {
        cerr << "error: " << e.what() << "\n";
        return 1;
    }
    catch(...) {
        cerr << "Exception of unknown type!\n";
    }

    if(baseport < 1024 || baseport > 65533) {
        cerr << "baseport is not a valid high port\n";
        return 1;
    }

    SendToHardwareViaMAC* mac_tx = new SendToHardwareViaMAC("unknown",interface);
    ReceiveFromHardwareViaMAC* mac_rx = new ReceiveFromHardwareViaMAC("unknown", interface);

    std::thread thread_radio_config (configure_radio_thread, baseport);

    std::thread thread_tx_wlan (tx_wlan_thread, mac_tx, baseport+1);
    std::thread thread_rx_wlan (rx_wlan_thread, mac_rx, mac_tx, baseport+2, echo, true);

    thread_tx_wlan.join();
    thread_rx_wlan.join();
    thread_radio_config.join();

    return EXIT_SUCCESS;
}
