// This file is part of Lanradio.
//
// Lanradio is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Lanradio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Lanradio.  If not, see <http://www.gnu.org/licenses/>.
//
// Authors:
// Florian Klingler <klingler@ccs-labs.org>

#ifndef WORLD_LANRADIOCOMMUNICATION_H
#define WORLD_LANRADIOCOMMUNICATION_H

#include <omnetpp.h>
#include "veins/base/utils/FindModule.h"
#include <thread>
#include <map>
#include <mutex>
#include "veins/modules/messages/WaveShortMessage_m.h"
#include "veins/modules/messages/WaveServiceAdvertisement_m.h"
#include "veins/modules/messages/BasicSafetyMessage_m.h"
#include <netinet/ether.h>
#include <veins/modules/lanradio/Constants.h>
#include <veins/modules/lanradio/GeoNetHeaders.h>
#include <veins/modules/lanradio/packets/MessageUtils.h>
#include <veins/modules/lanradio/protobuf/build/wlan.pb.h>
#include <zmq.h>
#include <zmq.hpp>
#include <veins/modules/mac/lanradio/IntermediateMacInterface.h>
#include "veins/modules/messages/lanradio/CAMMessage_m.h"
#include "veins/modules/utility/Consts80211p.h"
#include "veins/modules/messages/lanradio/SPATMessage_m.h"

namespace Veins {
class LanradioCommunication : public cSimpleModule
{
	public:

        enum t_wlan_type {
            wlan_CAM, wlan_DENM, wlan_SPAT
        };

        struct wlan_msg {
            std::string src_mac;
            std::string dst_mac;
            std::string payload;
        };

		virtual ~LanradioCommunication();
		virtual void initialize(int stage);
		virtual void finish();

		//virtual void sendMessage(std::string msg);
		virtual void send_WLAN(void* data, t_wlan_type type, std::string srcmac, std::string dstmac);
		virtual void insert_wlan_msg_buffer(wlan_msg msg);

		virtual bool registerRadio(IntermediateMacInterface* mac);
		virtual int getBasePort() {
		    return baseport;
		}
		virtual std::string getHostname() {
		    return hostname;
		}

	protected:

		enum msg_types {
		    Lanradio_CHECK_MSG_BUFFER
		};

		virtual void handleMessage(cMessage *msg) override;


		zmq::context_t* mContext;
		zmq::socket_t* mPublisher;

		void activate_radio();
		void activate_lanradio();
		void deactivate_radio();
		std::string radio_send_command(const std::string cmd);

		virtual void process_message_buffer();
		virtual void received_WLAN(wlan_msg m);

		std::string radioconfig_zmq_hostname;
		int radioconfig_zmq_port;
		zmq::context_t* radioconfig_zmq_context;
		zmq::socket_t* radioconfig_zmq_socket;

		bool* receive_thread_running;
		std::thread* receive_thread;

		std::mutex mutex_message_buffer;

		std::map<simtime_t, wlan_msg>* wlan_message_buffer;

		cMessage* check_message_buffer;

		IntermediateMacInterface* intermediate_mac;
		int baseport;
		std::string hostname;
		std::string wlan_interface;
		bool use_lanradio;
		std::string txpower; // in mBm
		std::string bitrates;
		std::string wlan_ip;
		std::string freq; // in MHz
		std::string bandwidth; // in MHz
		bool lara_debug;

		virtual void fillGeoNetBTPheaderForCam(CAM_t* cam, int payloadLen, std::string srcmac);
		struct GeoNetworkAndBTPHeaderCAM mGeoBtpHdrForCam;

		virtual void fillGeoNetBTPheaderForSpat(int payloadLen, std::string srcmac);
		struct GeoNetworkAndBTPHeaderSPAT mGeoBtpHdrForSpat;
};
}

namespace Veins {
class LanradioCommunicationAccess
{
	public:
    LanradioCommunication* get() {
			return FindModule<LanradioCommunication*>::findGlobalModule();
		};
};
}

#endif
