// This file is part of Lanradio.
//
// Lanradio is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Lanradio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Lanradio.  If not, see <http://www.gnu.org/licenses/>.
//
// Authors:
// Florian Klingler <klingler@ccs-labs.org>

#ifndef ___INTERMEDIATEMAC_H_
#define ___INTERMEDIATEMAC_H_

#include <assert.h>
#include <omnetpp.h>
#include <queue>
#include <stdint.h>
#include <veins/modules/lanradio/LanradioCommunication.h>
#include <veins/modules/mac/ieee80211p/LanradioToMac1609_4Interface.h>
#include <veins/modules/mac/lanradio/IntermediateMacInterface.h>
#include <veins/modules/messages/lanradio/CAMMessage_m.h>
#include <veins/modules/messages/lanradio/SPATMessage_m.h>
#include "veins/base/modules/BaseLayer.h"
#include "veins/base/utils/FindModule.h"
#include "veins/modules/messages/Mac80211Pkt_m.h"
#include "veins/modules/messages/WaveShortMessage_m.h"
#include "veins/base/modules/BaseMacLayer.h"
#include "veins/modules/utility/ConstsPhy.h"
#include "veins/modules/phy/PhyLayer80211p.h"
#include "veins/modules/messages/Mac80211Pkt_m.h"
#include "veins/base/phyLayer/MacToPhyControlInfo.h"

using Veins::LanradioCommunication;
using Veins::LanradioCommunicationAccess;


class IntermediateMac : public BaseMacLayer, public IntermediateMacInterface {

	public:
		~IntermediateMac() { };
		void handleLanradioMessage(cMessage* msg);

	protected:
		/** @brief Initialization of the module and some variables.*/
		virtual void initialize(int);

		/** @brief Delete all dynamically allocated objects of the module.*/
		virtual void finish();

		/** @brief Handle messages from lower layer.*/
		virtual void handleLowerMsg(cMessage*);

		/** @brief Handle messages from upper layer.*/
		virtual void handleUpperMsg(cMessage*);

		/** @brief Handle control messages from upper layer.*/
		virtual void handleUpperControl(cMessage* msg);


		/** @brief Handle self messages such as timers.*/
		virtual void handleSelfMsg(cMessage*);

		/** @brief Handle control messages from lower layer.*/
		virtual void handleLowerControl(cMessage* msg);

	protected:
		LanradioCommunication* lara;

		bool send_to_lanradio;

		LanradioToMac1609_4Interface* mac;
};

#endif /* ___INTERMEDIATEMAC_H_*/
